package PScm::Closure;

use strict;
use warnings;
use base qw(PScm);

sub new {
    my ($class, $args, $body, $env) = @_;

    bless {
        args => $args,
        body => $body,
        env  => $env,
    }, $class;
}

sub _apply {
    my ($self, $args, $cont) = @_;

    my $extended_env =
      $self->{env}->ExtendUnevaluated($self->{args}, $args);
    $self->{body}->Eval($extended_env, $cont);
}

sub as_string {
    my ($self) = @_;
    return PScm::Expr::List->new(
        $self->_symbol,
        $self->{args},
        $self->{body}
    )->as_string;
}

################################
package PScm::Closure::Function;

use base qw(PScm::Closure);
use PScm::Continuation;

sub Apply {
    my ($self, $form, $env, $cont) = @_;

    $form->map_eval(
        $env,
        cont {
            my ($evaluated_args) = @_;
            $self->_apply($evaluated_args, $cont);
        }
    );
}

sub _symbol {
    PScm::Expr::Symbol->new('lambda');
}

#############################
package PScm::Closure::Macro;

use base qw(PScm::Closure);
use PScm::Continuation;

sub Apply {
    my ($self, $form, $env, $cont) = @_;

    $self->_apply(
        $form,
        cont {
            my ($new_form) = @_;
            $new_form->Eval($env, $cont);
        }
    );
}

sub _symbol {
    PScm::Expr::Symbol->new('macro');
}

1;
