package PScm::SpecialForm;

use strict;
use warnings;
use base qw(PScm::Expr);

###############################
package PScm::SpecialForm::Let;

use base qw(PScm::SpecialForm);
use PScm::Continuation;

sub Apply {
    my ($self, $form, $env, $cont) = @_;

    my ($symbols, $values, $body) = $self->UnPack($form);

    $env->Extend(
        $symbols, $values,
        cont {
            my ($newenv) = @_;
            $body->Eval($newenv, $cont);
        }
    );
}

sub UnPack {
    my ($self, $form) = @_;

    my ($bindings, $body) = $form->value;
    my ($symbols, $values) = $self->unpack_bindings($bindings);
    return ($symbols, $values, $body);
}

sub unpack_bindings {
    my ($self, $bindings) = @_;
    if ($bindings->is_null) {
        my $null = new PScm::Expr::List::Null();
        return ($null, $null);
    } else {
        my ($symbols, $values) =
            $self->unpack_bindings($bindings->rest);
        return (
            PScm::Expr::List->Cons($bindings->first->first,
                                   $symbols),
            PScm::Expr::List->Cons($bindings->first->rest->first,
                                   $values)
        );
    }
}

##################################
package PScm::SpecialForm::LetRec;

use base qw(PScm::SpecialForm::Let);
use PScm::Continuation;

sub Apply {
    my ($self, $form, $env, $cont) = @_;

    my ($symbols, $values, $body) = $self->UnPack($form);

    $env->ExtendRecursively(
        $symbols, $values,
        cont {
            my ($extended_env) = @_;
            $body->Eval($extended_env, $cont);
        }
    );
}

###################################
package PScm::SpecialForm::LetStar;

use base qw(PScm::SpecialForm::Let);
use PScm::Continuation;

sub Apply {
    my ($self, $form, $env, $cont) = @_;

    my ($symbols, $values, $body) = $self->UnPack($form);

    $env->ExtendIteratively(
        $symbols, $values,
        cont {
            my ($extended_env) = @_;
            $body->Eval($extended_env, $cont);
        }
    );
}

##############################
package PScm::SpecialForm::If;

use base qw(PScm::SpecialForm);
use PScm::Continuation;

sub Apply {
    my ($self, $form, $env, $cont) = @_;

    my $condition = $form->first;
    my $true_branch = $form->rest->first;
    my $false_branch = $form->rest->rest->first;

    $condition->Eval(
        $env,
        cont {
            my ($result) = @_;
            if ($result->isTrue) {
                $true_branch->Eval($env, $cont);
            } else {
                $false_branch->Eval($env, $cont);
            }
        }
    );
}

################################
package PScm::SpecialForm::Eval;

use base qw(PScm::SpecialForm);
use PScm::Continuation;

sub Apply {
    my ($self, $form, $env, $cont) = @_;
    $form->first()->Eval(
        $env,
        cont {
            my ($result) = @_;
            $result->Eval($env, $cont);
        }
    );
}

##################################
package PScm::SpecialForm::Lambda;

use base qw(PScm::SpecialForm);
use PScm::Closure;

sub Apply {
    my ($self, $form, $env, $cont) = @_;

    my ($args, $body) = $form->value;

    $cont->Cont(PScm::Closure::Function->new($args, $body, $env));
}

##################################
package PScm::SpecialForm::CallCC;

use base qw(PScm::SpecialForm);
use PScm::Closure;
use PScm::Continuation;

sub Apply {
    my ($self, $form, $env, $cont) = @_;

    $form->first->Eval(
        $env,
        cont {
            my ($closure) = @_;
            $closure->Apply(new PScm::Expr::List($cont),
                $env, $cont);
        }
    );
}

##################################
package PScm::SpecialForm::Macro;

use base qw(PScm::SpecialForm);
use PScm::Closure;

sub Apply {
    my ($self, $form, $env, $cont) = @_;
    my ($args, $body) = $form->value;
    $cont->Cont(PScm::Closure::Macro->new($args, $body, $env));
}

#################################
package PScm::SpecialForm::Quote;

use base qw(PScm::SpecialForm);

sub Apply {
    my ($self, $form, $env, $cont) = @_;
    $form->first->Quote($env, $cont);
}

1;
